from .io import write_file, load_file

def sort_result(args):
    """Sort a result by (OR):
    - ok
    - status code
    - latency
    - errors
    """
    result = load_file(args.filename, format=args.in_format if args.in_format is not None else args.format)

    sorted_result = {}
    for url in result:

        if (args.ok is result[url].get("ok", None) and args.ok is not None) or (args.status_code == result[url].get("status_code", None) and args.status_code is not None) or (args.latency is not None and result[url]["latency"] > args.latency) or (args.error and result[url].get("error", None) is not None):
            sorted_result[url] = result[url]
        else:
            sorted_result[url] = {}

    write_file(sorted_result, format=args.out_format if args.out_format is not None else args.format)



from .io import load_file, write_file

def diff_results(args):
    filename_a = args.filename_a
    filename_b = args.filename_b

    if filename_a == "-":
        raise Exception("Only one file can be stdin")


    result_a = load_file(filename_a, format=args.format if args.in_format is None else  args.in_format)
    result_b = load_file(filename_b, format=args.format if args.in_format is None else args.in_format)

    diff = {}

    for url in result_a:
        diff[url] = { field: take_diff(result_a[url][field], result_b[url][field]) for field in result_a[url] }

    write_file(diff, format=args.format if args.out_format is None else args.out_format)

def take_diff(a, b):
    print(f"Diffing {a} {b}")
    if type(a) == list:
        return a > b
    elif b is None:
        return a
    elif a is None:
        return b
    elif type(a) is str and type(b) is str:
        return a > b
    else:
        return a - b



from html.parser import HTMLParser
from urllib.parse import urlparse

TAGS = ["link", "img", "video", "script", "a"]
ATTRS = ["href", "src"]

class Parser(HTMLParser):
    def handle_starttag(self, tag, attrs):
        if tag in TAGS:
            url = list(filter(lambda l: l[0] in ATTRS, attrs))

            if len(url) == 1:
                self._urls.add(url[0][1].strip())

    def __init__(self):
        super().__init__()
        self._urls = set()

def handle_relative_link(u, domain="", proto="http"):

        parsed = urlparse(u)

        if parsed._replace(netloc="").geturl() == u:

            if not u.startswith("/"):
                u = "/" + u

            u = domain + u

        parsed = urlparse(u)

        if parsed._replace(scheme="").geturl() == u:
            u = proto + "://" + u

        return u

def extract_links(text, domain="", proto="http"):
    parser = Parser()
    parser.feed(text)

    result = parser._urls

    result = { r for r in result if not r.startswith(("mailto", "e-name")) }

    result = { handle_relative_link(u, domain=domain, proto=proto) for u in result }

    return result



import sys
import json
import yaml
import csv

def or_none(fn):
    return lambda x: fn(x) if x != "" else None

CSV_FIELDS = {
    "url": or_none(lambda u: u[1:-1]),
    "ok": or_none(lambda o: o == "True"),
    "status_code": or_none(int),
    "latency": or_none(float),
    "occurances": or_none(lambda o: json.loads(o.replace("'", '"'))),
    "error": or_none(str)
}

def load_file(filename, format = "json"):
    """
    Load a file and parse it according to the format
    """
    def open_file_or_stdin(filename):
        if filename == "-":
            return sys.stdin
        else:
            return open(filename)

    with open_file_or_stdin(filename) as f:

        if format == "json":
            return json.load(f)
        elif format in ["yaml", "yml"]:
            return yaml.safe_load(f)
        elif format == "csv":
            reader = csv.DictReader(f)
            result = {}
            for row in reader:
                result[row["url"]] = { r: CSV_FIELDS[r](row[r]) for r in row if r != "url" }
            return result

def write_file(result, file_handler=sys.stdout, format = "json"):
    """
    Write a result to a file (stdout by default) according to the
    format.
    """
    if format == "json":
        return json.dump(result, file_handler)
    elif format in ["yml", "yaml"]:
        return yaml.dump(result, file_handler)
    elif format == "csv":
        fieldnames = CSV_FIELDS.keys()

        writer = csv.DictWriter(file_handler, fieldnames=fieldnames)

        writer.writeheader()

        for url in result:
            writer.writerow({ field: result[url].get(field,None) for field in result[url] } | { "url" : url })



#!/bin/env python3
import os
import sys
import argparse
from .extract_links import extract_links
from .check import check_locations
from .sort_result import sort_result
from .diff_results import diff_results
from .visualize import visualize
from urllib.parse import urlparse

def link(subparsers):
    links_parser = subparsers.add_parser("link", help="Tools to maintain the links on a website")

    formats = ["csv", "yaml", "yml", "json"]
    links_parser.add_argument("--format", "-f", help="Format to expect for input and output (is overwritten by in-format and out-format", choices=formats, default="json")
    links_parser.add_argument("--in-format", "-in-f", help="Format to expect from input only", choices=formats)
    links_parser.add_argument("--out-format", "-out-f", help="Format to use for output only", choices=formats)

    subparsers = links_parser.add_subparsers()

    # check
    check_parser = subparsers.add_parser("check", help="Check all the locations in the locations for their reachability, latency and status")
    check_parser.add_argument("locations", help="File of the locations to check")
    check_parser.add_argument("--num-processes", "-j",
                              type=int,
                              default=6,
                              help="Number of processes to run in parallel")
    check_parser.add_argument("--estimate-total", "-t",
                              type=int,
                              default=None,
                              help="Estimated number of total sites that'll be checked. Useful for nice progress bars")
    check_parser.set_defaults(func=check_locations)

    # sort
    sort_parser = subparsers.add_parser("sort", help="Sort the results from check according to some options")
    sort_parser.add_argument("filename",
                             help="Filename of the result",
                             default="-")
    sort_parser.add_argument("--ok",
                             action=argparse.BooleanOptionalAction,
                             help="Only return entries where the status was ok")
    sort_parser.add_argument("--status-code", "-s",
                             type=int,
                             help="Only return entries where the status code was this")
    sort_parser.add_argument("--latency", "-l",
                             type=float,
                             help="Only return entries where the latency was larger than this")
    sort_parser.add_argument("--error", "-e",
                             action=argparse.BooleanOptionalAction,
                             help="Only return entries where an error has occured")
    sort_parser.set_defaults(func=sort_result)

    # diff
    diff_parser = subparsers.add_parser("diff", help="Take the difference between two check results")
    diff_parser.add_argument("filename_a",
                             help="Filename of the first file. Cannot be - for stdin.")
    diff_parser.add_argument("filename_b",
                             help="Filename of the second file. Can be - for stdin.")
    diff_parser.set_defaults(func=diff_results)

    # visualize
    visualize_parser = subparsers.add_parser("visualize", help="Visualize the results of a page as a graph stored in a static HTML.")
    visualize_parser.add_argument("filename",
                                  default="-",
                                  help="Filename of the result")
    visualize_parser.add_argument("--output-filename", "-o",
                                  help="Filename to store the static HTML File in")
    visualize_parser.set_defaults(func=visualize)



from pyvis.network import Network
from urllib.parse import urlparse
from .io import load_file, write_file
from tqdm import tqdm

def visualize(args):
    result = load_file(args.filename, format=args.format if args.in_format is None else args.in_format)

    net = Network()

    domains = {
        "Locations File": {
            "urls": set(),
            "occured":set(),
            "occurances":set(),
            "ok":True,
            "error":True,
        }
    }

    with tqdm(total=3*len(result)) as progress:

        for url in result:
            domain = domain_from_url(url)

            if domain not in domains:
                domains[domain] = {
                    "urls": set(),
                    "occured" : set(),
                    "occurances": set(),
                    "ok": True,
                    "error": True,
                }

            domains[domain]["urls"].add(url)
            domains[domain]["occurances"] |= set(result[url]["occurances"]) if result[url]["occurances"] is not None else set()
            domains[domain]["ok"] = domains[domain]["ok"] and result[url].get("ok", False)
            domains[domain]["error"] = domains[domain]["error"] and result[url].get("error", False)

            if result[url]["occurances"] is None:
                continue

            for link in result[url]["occurances"]:
                domains[domain_from_url(url)]["occured"].add(url)

            progress.update(1)

        for domain in domains:
            net.add_node(domain,
                        title="\n".join(filter(lambda u: not result[u].get("ok", False) or not result[u].get("error", True),  domains[domain]["urls"])),
                        mass=len(domains[domain]["occured"]),
                        shape=node_shape(domain),
                        color=node_color(domains[domain]))

            progress.update(1)

        for domain in domains:
            for link in domains[domain]["occurances"]:
                net.add_edge(domain, domain_from_url(link))

            progress.update(1)

    net.toggle_physics(True)
    net.show(args.output_filename)

def domain_from_url(url):
    if url == "Locations File":
        return url

    return urlparse(url).netloc

def node_shape(domain):
    if "Locations File" == domain:
        return "box"
    else:
        return "ellipse"

def node_color(result):
    if result["error"]:
        return "red"
    elif result["ok"]:
        return "green"
    else:
        return "yellow"



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="maintain-website-tool",
    version="0.0.4",
    author="CSDUMMI",
    author_email="csdummi.misquality@simplelogin.co",
    description="Maintain a Website. Includes link checker.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://codeberg.org/developers/maintain-website-tool",
    project_urls={
        "Bug Tracker": "https://codeberg.org/developers/maintain-website-tool/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    package_dir={"":"src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "maintain-website-tool=maintain_website_tool.main:main",
        ]
    }
)



import multiprocessing
from urllib.parse import urlparse
from .io import write_file, load_file
from .extract_links import extract_links
from tqdm import tqdm
import time
import requests
import sys

def lookup_urls(locations, processes=6, estimate_total=None):
    """Run a manager to hold:
    - a set (as dict) of all unchecked urls,
    - a dict of urls to responses or exceptions occuring when getting those URLs
    - a dict of urls to a list of urls where that url occured in the HTML Plain Text.

    Then start processes (defaults to 6) to pop urls from the unchecked
    set, fetch the contents at that url and if possible extract links from
    those contents and add them to unchecked - if they share a domain with
    one of the locations provided as input.
    """

    domains = { urlparse(location).netloc for location in locations }

    urls = {}

    occurances = { location: ["Locations File"] for location in locations }

    with multiprocessing.Manager() as manager:

        unchecked = manager.dict({ location: None for location in locations }) # Acts as a set.

        urls = manager.dict()

        occurances = manager.list([(location, "Locations File") for location in locations])

        with multiprocessing.Pool(processes) as pool:
            loading = pool.map_async(check_url, [(unchecked, urls, occurances, domains, id) for id in range(processes)])

            with tqdm(unit="sites", total=estimate_total) as progress_checked:
                with tqdm(unit="sites", total=estimate_total) as progress_total:
                    progress_checked.set_description_str("Checked URLs")
                    progress_total.set_description_str("Total URLs found")

                    dt = 0.1
                    urls_len = len(urls)
                    total = len(unchecked) + len(urls)

                    while not loading.ready():
                        progress_checked.update(len(urls) - urls_len)
                        progress_total.update(len(unchecked) + len(urls) - total)

                        urls_len = len(urls)
                        total = len(unchecked) + urls_len

                        time.sleep(dt)

        occurances_ = {}

        for (link, origin) in occurances:
            if occurances_.get(link, None) is None:
                occurances_[link] = set()

            occurances_[link].add(origin)

        occurances = occurances_

        result = {}

        for url in urls:
            result[url] = urls[url].copy()

            response = result[url]["response"]
            del result[url]["response"]

            if isinstance(response, Exception):
                result[url]["error"] = str(response.__cause__)

            else:
                result[url]["ok"] = response.ok
                result[url]["status_code"] = response.status_code
                result[url]["occurances"] = list(occurances[url])

        return result

def check_url(arg):
    unchecked, urls, occurances, domains, id = arg

    while len(unchecked) != 0:
        try:
            url,_ = unchecked.popitem()

            before = time.time()
            response = requests.get(url)
            latency = time.time() - before

            urls[url] = {
                "response": response,
                "latency": latency,
            }

        except KeyError:
            continue
        except requests.exceptions.RequestException as e:
            urls[url] = {
                "response": e,
                "latency": time.time() - before
            }

        else:
            if not response.headers.get("Content-Type", default = "text/plain").startswith("text/html"):
                continue

            parsed = urlparse(url)

            if parsed.netloc in domains:
                links = extract_links(response.text, domain=parsed.netloc)

                for link in links:
                    occurances.append((link, url))

                    if link not in urls:
                        unchecked[link] = None

def check_locations(args):
    file_handler = open(args.locations) if args.locations not in ["", "-"] else sys.stdin
    locations = file_handler.read().strip().split("\n")

    result = lookup_urls(locations, processes = args.num_processes, estimate_total = args.estimate_total)

    write_file(result, format=args.out_format if args.out_format is not None else args.format)



#!/bin/env python3
from link import link
import argparse

def main():
    parser = argparse.ArgumentParser(prog="maintain-website-tool")
    parser.set_defaults(func=lambda args: parser.print_usage())

    subparsers = parser.add_subparsers()

    # Init subcommands
    link(subparsers)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()
